@@ -1,9 +1,10 @@
-use std::fmt;
-use std::io::{self, BufRead, BufReader, Cursor, ErrorKind, Read, Write};
+use log::debug;
+use std::io::{self, BufRead, BufReader, Read, Write};
 use std::net::SocketAddr;
 use std::net::TcpStream;
 use std::time::Duration;
 use std::time::Instant;
+use std::{fmt, io::Cursor};
 
 use chunked_transfer::Decoder as ChunkDecoder;
 
@@ -14,25 +15,22 @@ use rustls::StreamOwned;
 #[cfg(feature = "socks-proxy")]
 use socks::{TargetAddr, ToTargetAddr};
 
-#[cfg(feature = "native-tls")]
-use native_tls::{HandshakeError, TlsStream};
-
-use crate::proxy::Proto;
 use crate::proxy::Proxy;
+use crate::{error::Error, proxy::Proto};
 
-use crate::error::Error;
+use crate::error::ErrorKind;
 use crate::unit::Unit;
 
+pub(crate) struct Stream {
+    inner: BufReader<Inner>,
+}
+
 #[allow(clippy::large_enum_variant)]
-pub enum Stream {
-    Http(BufReader<TcpStream>),
-    #[cfg(all(feature = "tls", not(feature = "native-tls")))]
-    Https(BufReader<rustls::StreamOwned<rustls::ClientSession, TcpStream>>),
-    #[cfg(all(feature = "native-tls", not(feature = "tls")))]
-    Https(BufReader<TlsStream<TcpStream>>),
-    Cursor(Cursor<Vec<u8>>),
-    #[cfg(test)]
-    Test(Box<dyn BufRead + Send + Sync>, Vec<u8>),
+enum Inner {
+    Http(TcpStream),
+    #[cfg(feature = "tls")]
+    Https(rustls::StreamOwned<rustls::ClientSession, TcpStream>),
+    Test(Box<dyn Read + Send + Sync>, Vec<u8>),
 }
 
 // DeadlineStream wraps a stream such that read() will return an error
@@ -40,7 +38,7 @@ pub enum Stream {
 // TcpStream to ensure read() doesn't block beyond the deadline.
 // When the From trait is used to turn a DeadlineStream back into a
 // Stream (by PoolReturningRead), the timeouts are removed.
-pub struct DeadlineStream {
+pub(crate) struct DeadlineStream {
     stream: Stream,
     deadline: Option<Instant>,
 }
@@ -57,8 +55,8 @@ impl From<DeadlineStream> for Stream {
     }
 }
 
-impl Read for DeadlineStream {
-    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
+impl BufRead for DeadlineStream {
+    fn fill_buf(&mut self) -> io::Result<&[u8]> {
         if let Some(deadline) = self.deadline {
             let timeout = time_until_deadline(deadline)?;
             if let Some(socket) = self.stream.socket() {
@@ -66,17 +64,36 @@ impl Read for DeadlineStream {
                 socket.set_write_timeout(Some(timeout))?;
             }
         }
-        self.stream.read(buf).map_err(|e| {
+        self.stream.fill_buf().map_err(|e| {
             // On unix-y platforms set_read_timeout and set_write_timeout
             // causes ErrorKind::WouldBlock instead of ErrorKind::TimedOut.
             // Since the socket most definitely not set_nonblocking(true),
             // we can safely normalize WouldBlock to TimedOut
-            if e.kind() == ErrorKind::WouldBlock {
+            if e.kind() == io::ErrorKind::WouldBlock {
                 return io_err_timeout("timed out reading response".to_string());
             }
             e
         })
     }
+
+    fn consume(&mut self, amt: usize) {
+        self.stream.consume(amt)
+    }
+}
+
+impl Read for DeadlineStream {
+    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
+        // All reads on a DeadlineStream use the BufRead impl. This ensures
+        // that we have a chance to set the correct timeout before each recv
+        // syscall.
+        // Copied from the BufReader implementation of `read()`.
+        let nread = {
+            let mut rem = self.fill_buf()?;
+            rem.read(buf)?
+        };
+        self.consume(nread);
+        Ok(nread)
+    }
 }
 
 // If the deadline is in the future, return the remaining time until
@@ -90,30 +107,45 @@ fn time_until_deadline(deadline: Instant) -> io::Result<Duration> {
 }
 
 pub(crate) fn io_err_timeout(error: String) -> io::Error {
-    io::Error::new(ErrorKind::TimedOut, error)
+    io::Error::new(io::ErrorKind::TimedOut, error)
 }
 
 impl fmt::Debug for Stream {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(
-            f,
-            "Stream[{}]",
-            match self {
-                Stream::Http(_) => "http",
-                #[cfg(any(
-                    all(feature = "tls", not(feature = "native-tls")),
-                    all(feature = "native-tls", not(feature = "tls")),
-                ))]
-                Stream::Https(_) => "https",
-                Stream::Cursor(_) => "cursor",
-                #[cfg(test)]
-                Stream::Test(_, _) => "test",
-            }
-        )
+        match self.inner.get_ref() {
+            Inner::Http(tcpstream) => write!(f, "{:?}", tcpstream),
+            #[cfg(feature = "tls")]
+            Inner::Https(tlsstream) => write!(f, "{:?}", tlsstream.get_ref()),
+            Inner::Test(_, _) => write!(f, "Stream(Test)"),
+        }
     }
 }
 
 impl Stream {
+    fn logged_create(stream: Stream) -> Stream {
+        debug!("created stream: {:?}", stream);
+        stream
+    }
+
+    pub(crate) fn from_vec(v: Vec<u8>) -> Stream {
+        Stream::logged_create(Stream {
+            inner: BufReader::new(Inner::Test(Box::new(Cursor::new(v)), vec![])),
+        })
+    }
+
+    fn from_tcp_stream(t: TcpStream) -> Stream {
+        Stream::logged_create(Stream {
+            inner: BufReader::new(Inner::Http(t)),
+        })
+    }
+
+    #[cfg(feature = "tls")]
+    fn from_tls_stream(t: StreamOwned<ClientSession, TcpStream>) -> Stream {
+        Stream::logged_create(Stream {
+            inner: BufReader::new(Inner::Https(t)),
+        })
+    }
+
     // Check if the server has closed a stream by performing a one-byte
     // non-blocking read. If this returns EOF, the server has closed the
     // connection: return true. If this returns a successful read, there are
@@ -133,7 +165,7 @@ impl Stream {
 
         let result = match stream.peek(&mut buf) {
             Ok(n) => {
-                log::debug!(
+                debug!(
                     "peek on reused connection returned {}, not WouldBlock; discarding",
                     n
                 );
@@ -154,13 +186,10 @@ impl Stream {
         }
     }
     pub fn is_poolable(&self) -> bool {
-        match self {
-            Stream::Http(_) => true,
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(_) => true,
+        match self.inner.get_ref() {
+            Inner::Http(_) => true,
+            #[cfg(feature = "tls")]
+            Inner::Https(_) => true,
             _ => false,
         }
     }
@@ -177,10 +206,10 @@ impl Stream {
     }
 
     pub(crate) fn socket(&self) -> Option<&TcpStream> {
-        match self {
-            Stream::Http(b) => Some(b.get_ref()),
+        match self.inner.get_ref() {
+            Inner::Http(b) => Some(b),
             #[cfg(feature = "tls")]
-            Stream::Https(b) => Some(&b.get_ref().sock),
+            Inner::Https(b) => Some(&b.get_ref()),
             _ => None,
         }
     }
@@ -195,56 +224,37 @@ impl Stream {
 
     #[cfg(test)]
     pub fn to_write_vec(&self) -> Vec<u8> {
-        match self {
-            Stream::Test(_, writer) => writer.clone(),
+        match self.inner.get_ref() {
+            Inner::Test(_, writer) => writer.clone(),
             _ => panic!("to_write_vec on non Test stream"),
         }
     }
 }
 
 impl Read for Stream {
+    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
+        self.inner.read(buf)
+    }
+}
+
+impl Read for Inner {
     fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
         match self {
-            Stream::Http(sock) => sock.read(buf),
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(stream) => read_https(stream, buf),
-            Stream::Cursor(read) => read.read(buf),
-            #[cfg(test)]
-            Stream::Test(reader, _) => reader.read(buf),
+            Inner::Http(sock) => sock.read(buf),
+            #[cfg(feature = "tls")]
+            Inner::Https(stream) => read_https(stream, buf),
+            Inner::Test(reader, _) => reader.read(buf),
         }
     }
 }
 
 impl BufRead for Stream {
     fn fill_buf(&mut self) -> io::Result<&[u8]> {
-        match self {
-            Stream::Http(r) => r.fill_buf(),
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(r) => r.fill_buf(),
-            Stream::Cursor(r) => r.fill_buf(),
-            #[cfg(test)]
-            Stream::Test(r, _) => r.fill_buf(),
-        }
+        self.inner.fill_buf()
     }
 
     fn consume(&mut self, amt: usize) {
-        match self {
-            Stream::Http(r) => r.consume(amt),
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(r) => r.consume(amt),
-            Stream::Cursor(r) => r.consume(amt),
-            #[cfg(test)]
-            Stream::Test(r, _) => r.consume(amt),
-        }
+        self.inner.consume(amt)
     }
 }
 
@@ -258,9 +268,9 @@ where
     }
 }
 
-#[cfg(all(feature = "tls", not(feature = "native-tls")))]
+#[cfg(feature = "tls")]
 fn read_https(
-    stream: &mut BufReader<StreamOwned<ClientSession, TcpStream>>,
+    stream: &mut StreamOwned<ClientSession, TcpStream>,
     buf: &mut [u8],
 ) -> io::Result<usize> {
     match stream.read(buf) {
@@ -270,19 +280,10 @@ fn read_https(
     }
 }
 
-#[cfg(all(feature = "native-tls", not(feature = "tls")))]
-fn read_https(stream: &mut BufReader<TlsStream<TcpStream>>, buf: &mut [u8]) -> io::Result<usize> {
-    match stream.read(buf) {
-        Ok(size) => Ok(size),
-        Err(ref e) if is_close_notify(e) => Ok(0),
-        Err(e) => Err(e),
-    }
-}
-
 #[allow(deprecated)]
-#[cfg(any(feature = "tls", feature = "native-tls"))]
+#[cfg(feature = "tls")]
 fn is_close_notify(e: &std::io::Error) -> bool {
-    if e.kind() != ErrorKind::ConnectionAborted {
+    if e.kind() != io::ErrorKind::ConnectionAborted {
         return false;
     }
 
@@ -297,46 +298,40 @@ fn is_close_notify(e: &std::io::Error) -> bool {
 
 impl Write for Stream {
     fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
-        match self {
-            Stream::Http(sock) => sock.get_mut().write(buf),
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(stream) => stream.get_mut().write(buf),
-            Stream::Cursor(_) => panic!("Write to read only stream"),
-            #[cfg(test)]
-            Stream::Test(_, writer) => writer.write(buf),
+        match self.inner.get_mut() {
+            Inner::Http(sock) => sock.write(buf),
+            #[cfg(feature = "tls")]
+            Inner::Https(stream) => stream.write(buf),
+            Inner::Test(_, writer) => writer.write(buf),
         }
     }
     fn flush(&mut self) -> io::Result<()> {
-        match self {
-            Stream::Http(sock) => sock.get_mut().flush(),
-            #[cfg(any(
-                all(feature = "tls", not(feature = "native-tls")),
-                all(feature = "native-tls", not(feature = "tls")),
-            ))]
-            Stream::Https(stream) => stream.get_mut().flush(),
-            Stream::Cursor(_) => panic!("Flush read only stream"),
-            #[cfg(test)]
-            Stream::Test(_, writer) => writer.flush(),
+        match self.inner.get_mut() {
+            Inner::Http(sock) => sock.flush(),
+            #[cfg(feature = "tls")]
+            Inner::Https(stream) => stream.flush(),
+            Inner::Test(_, writer) => writer.flush(),
         }
     }
 }
 
+impl Drop for Stream {
+    fn drop(&mut self) {
+        debug!("dropping stream: {:?}", self);
+    }
+}
+
 pub(crate) fn connect_http(unit: &Unit, hostname: &str) -> Result<Stream, Error> {
     //
     let port = unit.url.port().unwrap_or(80);
 
-    connect_host(unit, hostname, port)
-        .map(BufReader::new)
-        .map(Stream::Http)
+    connect_host(unit, hostname, port).map(Stream::from_tcp_stream)
 }
 
 #[cfg(all(feature = "tls", feature = "native-certs"))]
 fn configure_certs(config: &mut rustls::ClientConfig) {
     config.root_store =
-        rustls_native_certs::load_native_certs().expect("Could not load patform certs");
+        rustls_native_certs::load_native_certs().expect("Could not load platform certs");
 }
 
 #[cfg(all(feature = "tls", not(feature = "native-certs")))]
@@ -346,9 +341,10 @@ fn configure_certs(config: &mut rustls::ClientConfig) {
         .add_server_trust_anchors(&webpki_roots::TLS_SERVER_ROOTS);
 }
 
-#[cfg(all(feature = "tls", not(feature = "native-tls")))]
+#[cfg(feature = "tls")]
 pub(crate) fn connect_https(unit: &Unit, hostname: &str) -> Result<Stream, Error> {
     use once_cell::sync::Lazy;
+    use rustls::Session;
     use std::sync::Arc;
 
     static TLS_CONF: Lazy<Arc<rustls::ClientConfig>> = Lazy::new(|| {
@@ -360,52 +356,32 @@ pub(crate) fn connect_https(unit: &Unit, hostname: &str) -> Result<Stream, Error
     let port = unit.url.port().unwrap_or(443);
 
     let sni = webpki::DNSNameRef::try_from_ascii_str(hostname)
-        .map_err(|err| Error::DnsFailed(err.to_string()))?;
+        .map_err(|err| ErrorKind::Dns.new().src(err))?;
     let tls_conf: &Arc<rustls::ClientConfig> = unit
-        .req
+        .agent
+        .config
         .tls_config
         .as_ref()
         .map(|c| &c.0)
         .unwrap_or(&*TLS_CONF);
-    let sess = rustls::ClientSession::new(&tls_conf, sni);
-
-    let sock = connect_host(unit, hostname, port)?;
+    let mut sock = connect_host(unit, hostname, port)?;
+    let mut sess = rustls::ClientSession::new(&tls_conf, sni);
 
+    sess.complete_io(&mut sock)
+        .map_err(|err| ErrorKind::ConnectionFailed.new().src(err))?;
     let stream = rustls::StreamOwned::new(sess, sock);
 
-    Ok(Stream::Https(BufReader::new(stream)))
-}
-
-#[cfg(all(feature = "native-tls", not(feature = "tls")))]
-pub(crate) fn connect_https(unit: &Unit, hostname: &str) -> Result<Stream, Error> {
-    use std::sync::Arc;
-
-    let port = unit.url.port().unwrap_or(443);
-    let sock = connect_host(unit, hostname, port)?;
-
-    let tls_connector: Arc<native_tls::TlsConnector> = match &unit.req.tls_connector {
-        Some(connector) => connector.0.clone(),
-        None => Arc::new(native_tls::TlsConnector::new().map_err(|e| Error::TlsError(e))?),
-    };
-    let stream = tls_connector
-        .connect(&hostname.trim_matches(|c| c == '[' || c == ']'), sock)
-        .map_err(|e| match e {
-            HandshakeError::Failure(err) => Error::TlsError(err),
-            // The only other possibility is WouldBlock. Since we don't
-            // handle retries of WouldBlock, turn it into a generic error.
-            _ => Error::ConnectionFailed("TLS handshake unexpected error".to_string()),
-        })?;
-
-    Ok(Stream::Https(BufReader::new(stream)))
+    Ok(Stream::from_tls_stream(stream))
 }
 
 pub(crate) fn connect_host(unit: &Unit, hostname: &str, port: u16) -> Result<TcpStream, Error> {
-    let deadline: Option<Instant> = if unit.req.timeout_connect > 0 {
-        Instant::now().checked_add(Duration::from_millis(unit.req.timeout_connect))
-    } else {
-        unit.deadline
-    };
-    let proxy: Option<Proxy> = unit.req.proxy();
+    let connect_deadline: Option<Instant> =
+        if let Some(timeout_connect) = unit.agent.config.timeout_connect {
+            Instant::now().checked_add(timeout_connect)
+        } else {
+            unit.deadline
+        };
+    let proxy: Option<Proxy> = unit.agent.config.proxy.clone();
     let netloc = match proxy {
         Some(ref proxy) => format!("{}:{}", proxy.server, proxy.port),
         None => format!("{}:{}", hostname, port),
@@ -415,10 +391,10 @@ pub(crate) fn connect_host(unit: &Unit, hostname: &str, port: u16) -> Result<Tcp
     let sock_addrs = unit
         .resolver()
         .resolve(&netloc)
-        .map_err(|e| Error::DnsFailed(format!("{}", e)))?;
+        .map_err(|e| ErrorKind::Dns.new().src(e))?;
 
     if sock_addrs.is_empty() {
-        return Err(Error::DnsFailed(format!("No ip address for {}", hostname)));
+        return Err(ErrorKind::Dns.msg(&format!("No ip address for {}", hostname)));
     }
 
     let proto = if let Some(ref proxy) = proxy {
@@ -432,17 +408,18 @@ pub(crate) fn connect_host(unit: &Unit, hostname: &str, port: u16) -> Result<Tcp
     // Find the first sock_addr that accepts a connection
     for sock_addr in sock_addrs {
         // ensure connect timeout or overall timeout aren't yet hit.
-        let timeout = match deadline {
+        let timeout = match connect_deadline {
             Some(deadline) => Some(time_until_deadline(deadline)?),
             None => None,
         };
 
+        debug!("connecting to {} at {}", netloc, &sock_addr);
         // connect with a configured timeout.
         let stream = if Some(Proto::SOCKS5) == proto {
             connect_socks5(
                 &unit,
                 proxy.clone().unwrap(),
-                deadline,
+                connect_deadline,
                 sock_addr,
                 hostname,
                 port,
@@ -463,21 +440,22 @@ pub(crate) fn connect_host(unit: &Unit, hostname: &str, port: u16) -> Result<Tcp
 
     let mut stream = if let Some(stream) = any_stream {
         stream
+    } else if let Some(e) = any_err {
+        return Err(ErrorKind::ConnectionFailed.msg("Connect error").src(e));
     } else {
-        let err = Error::ConnectionFailed(format!("{}", any_err.expect("Connect error")));
-        return Err(err);
+        panic!("shouldn't happen: failed to connect to all IPs, but no error");
     };
 
-    if let Some(deadline) = deadline {
+    if let Some(deadline) = unit.deadline {
         stream.set_read_timeout(Some(time_until_deadline(deadline)?))?;
     } else {
-        stream.set_read_timeout(unit.req.timeout_read)?;
+        stream.set_read_timeout(unit.agent.config.timeout_read)?;
     }
 
-    if let Some(deadline) = deadline {
+    if let Some(deadline) = unit.deadline {
         stream.set_write_timeout(Some(time_until_deadline(deadline)?))?;
     } else {
-        stream.set_read_timeout(unit.req.timeout_read)?;
+        stream.set_write_timeout(unit.agent.config.timeout_write)?;
     }
 
     if proto == Some(Proto::HTTPConnect) {
@@ -512,11 +490,13 @@ fn socks5_local_nslookup(
     let addrs: Vec<SocketAddr> = unit
         .resolver()
         .resolve(&format!("{}:{}", hostname, port))
-        .map_err(|e| std::io::Error::new(ErrorKind::NotFound, format!("DNS failure: {}.", e)))?;
+        .map_err(|e| {
+            std::io::Error::new(io::ErrorKind::NotFound, format!("DNS failure: {}.", e))
+        })?;
 
     if addrs.is_empty() {
         return Err(std::io::Error::new(
-            ErrorKind::NotFound,
+            io::ErrorKind::NotFound,
             "DNS failure: no socket addrs found.",
         ));
     }
@@ -525,7 +505,7 @@ fn socks5_local_nslookup(
         Ok(addr) => Ok(addr),
         Err(err) => {
             return Err(std::io::Error::new(
-                ErrorKind::NotFound,
+                io::ErrorKind::NotFound,
                 format!("DNS failure: {}.", err),
             ))
         }
@@ -646,7 +626,7 @@ fn connect_socks5(
     _port: u16,
 ) -> Result<TcpStream, std::io::Error> {
     Err(std::io::Error::new(
-        ErrorKind::Other,
+        io::ErrorKind::Other,
         "SOCKS5 feature disabled.",
     ))
 }
@@ -659,10 +639,12 @@ pub(crate) fn connect_test(unit: &Unit) -> Result<Stream, Error> {
 
 #[cfg(not(test))]
 pub(crate) fn connect_test(unit: &Unit) -> Result<Stream, Error> {
-    Err(Error::UnknownScheme(unit.url.scheme().to_string()))
+    Err(ErrorKind::UnknownScheme.msg(&format!("unknown scheme '{}'", unit.url.scheme())))
 }
 
-#[cfg(not(any(feature = "tls", feature = "native-tls")))]
+#[cfg(not(feature = "tls"))]
 pub(crate) fn connect_https(unit: &Unit, _hostname: &str) -> Result<Stream, Error> {
-    Err(Error::UnknownScheme(unit.url.scheme().to_string()))
+    Err(ErrorKind::UnknownScheme
+        .msg("URL has 'https:' scheme but ureq was build without HTTP support")
+        .url(unit.url.clone()))
 }
